from ledger.client import LedgerClient, OperationType
from decimal import Decimal
from enum import Enum

import logging

log = logging.getLogger(__name__)


class ErrorCodes(Enum):
    FETCH_ORDER_STATUS_ERROR = 'Error while fetching fill statuses'
    INSUFFICIENT_FUNDS = 'Insufficient funds'
    FILL_REQUEST_SEQUENCE_ERROR = 'Fill request sequence error'


class Accountant:
    """
    Accountant binds together 2 concepts - BankAccount & Ledger
    Ledger is the source of truth for the Bank
    Ledger ALWAYS reflect the true state of the OutsideWorld
    Accountant MUST manage BankAccount's books inside Ledger properly
    Accountant stores no state
    """
    DEFAULT_CASH_BOOK_ID = {'name': 'cash_book', 'id': '1'}

    def __init__(self, bank_account, ledger_host, ledger_read_host=None):
        self.bank_account = bank_account
        self.ledger = LedgerClient(ledger_endpoint=ledger_host, ledger_read_only_endpoint=ledger_read_host)

        log.info("---------Initialising-------------")
        log.info(f"Bank Account Number:  {self.bank_account['account_id']}")
        log.info(f"Cash Book:  {self.DEFAULT_CASH_BOOK_ID}")

    def get_book(self, timestamp: int = None):
        log.info("---------Get Book----------")
        blocked_balance_book = self.ledger.get_book(book_id=self.bank_account['ledger_books']['blocked_balance']['id'], timestamp=timestamp)['balances']
        main_balance_book = self.ledger.get_book(book_id=self.bank_account['ledger_books']['main_balance']['id'], timestamp=timestamp)['balances']
        blocked_balance_deposit_book = self.ledger.get_book_balances(book_id=self.bank_account['ledger_books']['blocked_balance']['id'], metadata_filter={'operation': 'DEPOSIT'}, timestamp=timestamp)
        blocked_balance_withdraw_book = self.ledger.get_book_balances(book_id=self.bank_account['ledger_books']['blocked_balance']['id'], metadata_filter={'operation': 'WITHDRAW'}, timestamp=timestamp)
        blocked_balance_order_book = self.ledger.get_book_balances(book_id=self.bank_account['ledger_books']['blocked_balance']['id'], metadata_filter={'operation': 'ORDER'}, timestamp=timestamp)
        book_balance = []
        block_book_currency_list = [currency for currency, amount in blocked_balance_book.items()]
        main_book_currency_list =  [currency for currency, amount in main_balance_book.items()]
        all_currency_list = list(set(block_book_currency_list + main_book_currency_list))
        for currency in all_currency_list:
            if currency in main_balance_book:
                main_balance = main_balance_book[currency]
            else:
                main_balance = "0"
            if currency in blocked_balance_deposit_book:
                blocked_balance_deposit = blocked_balance_deposit_book[currency]
            else:
                blocked_balance_deposit = "0"
            if currency in blocked_balance_withdraw_book:
                blocked_balance_withdraw = blocked_balance_withdraw_book[currency]
            else:
                blocked_balance_withdraw = "0"
            if currency in blocked_balance_order_book:
                blocked_balance_order = blocked_balance_order_book[currency]
            else:
                blocked_balance_order = "0"
            balance = {
                'currency': currency,
                'blocked_balance_deposit': blocked_balance_deposit,
                'blocked_balance_withdraw': blocked_balance_withdraw,
                'blocked_balance_order': blocked_balance_order,
                'main_balance': main_balance
            }
            book_balance.append(dict(balance))
        return book_balance

    def get_book_balances(self, currency: str, metadata_filter: dict = {}, timestamp: int = None):
        log.info("---------Balance----------")
        blocked_deposit_balance = self.ledger.get_book_balances(
            book_id=self.bank_account['ledger_books']['blocked_balance']['id'],
            asset_id=currency, metadata_filter={**{'operation': 'DEPOSIT'}, **metadata_filter}, timestamp=timestamp)
        blocked_withdraw_balance = self.ledger.get_book_balances(
            book_id=self.bank_account['ledger_books']['blocked_balance']['id'],
            asset_id=currency, metadata_filter={**{'operation': 'WITHDRAW'}, **metadata_filter}, timestamp=timestamp)
        blocked_order_balance = self.ledger.get_book_balances(
            book_id=self.bank_account['ledger_books']['blocked_balance']['id'],
            asset_id=currency, metadata_filter={**{'operation': 'ORDER'}, **metadata_filter}, timestamp=timestamp)
        main_balance = self.ledger.get_book_balances(book_id=self.bank_account['ledger_books']['main_balance']['id'],
                                                     asset_id=currency, metadata_filter={**metadata_filter}, timestamp=timestamp)
        return {'currency': currency,
                'blocked_balance_deposit': blocked_deposit_balance.get(currency, "0"),
                'blocked_balance_withdraw': blocked_withdraw_balance.get(currency, "0"),
                'blocked_balance_order': blocked_order_balance.get(currency, "0"),
                'main_balance': main_balance.get(currency, "0")}

    def withdraw_block_amount(self, value: str, currency: str, memo: str, metadata: dict = {}):
        log.info("---------Block_Amount_Withdraw----------")
        operation = self.ledger.post_transfer(
            from_book_id=self.bank_account['ledger_books']['main_balance']['id'],
            to_book_id=self.bank_account['ledger_books']['blocked_balance']['id'], asset_id=currency, value=value,
            memo=memo, metadata={**{"operation": "WITHDRAW"}, **metadata})
        log.info(operation)
        return operation

    def deposit_block_amount(self, value: str, currency: str, memo: str, metadata: dict = {}):
        log.info("---------Block_Amount_Deposit----------")
        operation = self.ledger.post_transfer(
            from_book_id=self.DEFAULT_CASH_BOOK_ID['id'],
            to_book_id=self.bank_account['ledger_books']['blocked_balance']['id'], asset_id=currency, value=value,
            memo=memo,
            metadata={**{'operation': 'DEPOSIT'}, **metadata})
        log.info(operation)
        return operation

    def withdraw_unblock_amount(self, value: str, currency: str, memo: str, metadata: dict = {}):
        log.info("---------Unblock_Amount_Withdraw----------")
        operation = self.ledger.post_transfer(
            from_book_id=self.bank_account['ledger_books']['blocked_balance']['id'],
            to_book_id=self.bank_account['ledger_books']['main_balance']['id'], asset_id=currency, value=value,
            memo=memo,
            metadata={**{'operation': 'WITHDRAW'}, **metadata})
        log.info(operation)
        return operation

    def deposit_unblock_amount(self, value: str, currency: str, memo: str, metadata: dict = {}):
        log.info("---------Unblock_Amount_Deposit----------")
        operation = self.ledger.post_transfer(
            from_book_id=self.bank_account['ledger_books']['blocked_balance']['id'],
            to_book_id=self.DEFAULT_CASH_BOOK_ID['id'], asset_id=currency, value=value, memo=memo,
            metadata={**{'operation': 'DEPOSIT'}, **metadata})
        log.info(operation)
        return operation

    def withdraw_amount(self, value: str, currency: str, memo: str, metadata: dict = {}):
        log.info("---------Withdraw_amount----------")
        operation = self.ledger.post_transfer(
            from_book_id=self.bank_account['ledger_books']['blocked_balance']['id'],
            to_book_id=self.DEFAULT_CASH_BOOK_ID['id'],
            asset_id=currency, value=value, memo=memo, metadata={**{'operation': 'WITHDRAW'}, **metadata})
        log.info(operation)
        return operation

    def deposit_amount(self, value: str, currency: str, memo: str, metadata: dict = {}):
        log.info("---------Deposit_amount----------")
        operation = self.ledger.post_transfer(
            from_book_id=self.bank_account['ledger_books']['blocked_balance']['id'],
            to_book_id=self.bank_account['ledger_books'][
                'main_balance']['id'], asset_id=currency, value=value,
            memo=memo, metadata={**{'operation': 'DEPOSIT'}, **metadata})
        log.info(operation)
        return operation

    def block_trade_amount(self, from_currency: str, to_currency: str, from_amount: str, to_amount: str, memo: str, metadata: dict = {}):
        log.info("---------Block_Trade_Amount----------")
        withdraw_block_amount = self.withdraw_block_amount(value=from_amount, currency=from_currency, memo=memo + "_WITHDRAW_BLOCK", metadata=metadata)
        deposit_block_amount = self.deposit_block_amount(value=to_amount, currency=to_currency, memo=memo + "_DEPOSIT_BLOCK", metadata=metadata)
        operation = {
            'withdraw_block_amount': withdraw_block_amount,
            'deposit_block_amount': deposit_block_amount
        }
        log.info(operation)
        return operation

    def trade_amount(self, from_currency: str, to_currency: str, from_amount: str, to_amount: str, memo: str, metadata: dict = {}):
        log.info("---------Trade_Amount----------")
        withdraw_amount = self.withdraw_amount(value=from_amount, currency=from_currency, memo=memo + "_WITHDRAW", metadata=metadata)
        deposit_amount = self.deposit_amount(value=to_amount, currency=to_currency, memo=memo + "_DEPOSIT", metadata=metadata)
        operation = {
            'withdraw_block_amount': withdraw_amount,
            'deposit_block_amount': deposit_amount
        }
        log.info(operation)
        return operation

    def freeze_books(self):
        log.info("---------Freezing Ledger Books----------")
        blocked_book = self.ledger.freeze_book(self.bank_account['ledger_books']['blocked_balance']['id'])
        main_book = self.ledger.freeze_book(self.bank_account['ledger_books']['main_balance']['id'])
        log.info(blocked_book, main_book)
        return main_book, blocked_book

    def unfreeze_books(self):
        log.info("---------Unfreezing Ledger Books----------")
        blocked_book = self.ledger.unfreeze_book(self.bank_account['ledger_books']['blocked_balance']['id'])
        main_book = self.ledger.unfreeze_book(self.bank_account['ledger_books']['main_balance']['id'])
        log.info(blocked_book, main_book)
        return main_book, blocked_book

    @staticmethod
    def create_book(account_id: str, ledger_host):
        log.info("---------New_ledger-------------")
        ledger = LedgerClient(ledger_endpoint=ledger_host)
        block_book_id = ledger.create_book(name=f"{account_id}_block_book", min_balance="0")
        main_book_id = ledger.create_book(name=f"{account_id}_main_book", min_balance="0")
        new_books = {
            'main_balance': {'name': main_book_id['name'],
                             'id': main_book_id['id']},
            'blocked_balance': {'name': block_book_id['name'],
                                'id': block_book_id['id']}
        }
        log.info(new_books)
        return new_books

    def block_for_order(self, order_id: str, order_type: str, currency: str, value: str, metadata: dict = {}):
        log.info("---------Block_Limit_Order_Funds----------")

        memo = f"{order_id}_BLOCK_ORDER"

        metadata['order_id'] = order_id
        metadata['order_type'] = order_type
        metadata['operation'] = 'ORDER'
        metadata['from_currency'] = currency
        metadata['type'] = 'BLOCK'

        operation = self.ledger.post_transfer(
            from_book_id=self.bank_account['ledger_books']['main_balance']['id'],
            to_book_id=self.bank_account['ledger_books']['blocked_balance']['id'],
            asset_id=currency,
            value=value,
            memo=memo,
            metadata=metadata
        )

        log.info(operation)
        return operation

    def get_book_balances_bulk(self, fetch_details: list):
        query = []
        for fetch in fetch_details:
            query.append(
                {
                    'book_id': fetch['book_id'],
                    'metadata_filter': fetch['metadata_filter']
                }
            )
        balance_list = self.ledger.get_book_balances_bulk(query)
        return balance_list

    def unblock_for_order(self, order_id: str, order_type: str, metadata: dict = {}):
        log.info("---------Unblock_Limit_Order_Funds----------")

        memo = f"{order_id}_UNBLOCK_ORDER"

        block_book_order_balances = self.ledger.get_book_balances(
            book_id=self.bank_account['ledger_books']['blocked_balance']['id'],
            metadata_filter={
                "order_id": order_id
            }
        )

        if block_book_order_balances == {}:
            raise Exception(ErrorCodes.INSUFFICIENT_FUNDS.value)

        metadata['order_id'] = order_id
        metadata['order_type'] = order_type
        metadata['type'] = 'UNBLOCK'
        metadata['operation'] = 'ORDER'

        entries = []
        for asset, value in block_book_order_balances.items():
            entries.extend([
                {
                    "bookId": self.bank_account['ledger_books']['blocked_balance']['id'],
                    "assetId": asset,
                    "value": str(Decimal(value) * Decimal("-1"))
                },
                {
                    "bookId": self.bank_account['ledger_books']['main_balance']['id'],
                    "assetId": asset,
                    "value": str(Decimal(value))
                }
            ])

        operation = self.ledger.post_operation(OperationType.CUSTOM, entries, memo, metadata)
        log.info(operation)
        return operation

    def update_order_fill_status(self, order_id, order_type: str, from_currency: str, to_currency: str, from_amount: str, to_amount: str, memo: str, metadata: dict = {}):
        log.info("---------Update_Limit_Order_Funds----------")

        memo = f"{memo}_{order_id}_FILL_ORDER"

        [main_book_order_balances, block_book_order_balances, cash_book_order_balances] = self.get_book_balances_bulk([
            {
                'book_id': self.bank_account['ledger_books']['main_balance']['id'],
                'metadata_filter': {'order_id': order_id}
            },
            {
                'book_id': self.bank_account['ledger_books']['blocked_balance']['id'],
                'metadata_filter':{'order_id':order_id}
            },
            {
                'book_id': self.DEFAULT_CASH_BOOK_ID['id'],
                'metadata_filter': {'order_id': order_id}
            }
        ])

        if block_book_order_balances == {}:
            raise Exception(ErrorCodes.INSUFFICIENT_FUNDS.value)

        from_currency_balance = cash_book_order_balances.get(from_currency, "0")
        to_currency_balance = main_book_order_balances.get(to_currency, "0")

        if Decimal(to_amount) < Decimal(to_currency_balance) or Decimal(from_amount) < Decimal(from_currency_balance):
            raise Exception(ErrorCodes.FILL_REQUEST_SEQUENCE_ERROR.value)

        delta_from_amount = str(Decimal(from_amount) - Decimal(from_currency_balance))
        delta_to_amount = str(Decimal(to_amount) - Decimal(to_currency_balance))

        if Decimal(delta_from_amount) == Decimal("0") and Decimal(delta_to_amount) == Decimal("0"):
            return

        if Decimal(delta_from_amount) > Decimal(block_book_order_balances.get(from_currency, "0")):
            raise Exception(ErrorCodes.INSUFFICIENT_FUNDS.value)

        metadata['order_id'] = order_id
        metadata['order_type'] = order_type
        metadata['type'] = 'FILL'
        metadata['operation'] = 'ORDER'

        entries = [
            {
                "bookId": self.bank_account['ledger_books']['main_balance']['id'],
                "assetId": to_currency,
                "value": delta_to_amount
            },
            {
                "bookId": self.bank_account['ledger_books']['blocked_balance']['id'],
                "assetId": from_currency,
                "value": str(Decimal(delta_from_amount) * Decimal("-1"))
            },
            {
                "bookId": self.DEFAULT_CASH_BOOK_ID['id'],
                "assetId": from_currency,
                "value": delta_from_amount
            },
            {
                "bookId": self.DEFAULT_CASH_BOOK_ID['id'],
                "assetId": to_currency,
                "value": str(Decimal(delta_to_amount) * Decimal("-1"))
            }
        ]

        operation = self.ledger.post_operation(OperationType.CUSTOM, entries, memo, metadata)
        log.info(operation)
        return operation

    def trade_all_or_none(self, from_currency: str, to_currency: str, from_amount: str, to_amount: str, memo: str, metadata: dict = {}):
       # operation in metadata would be trade now, {**metadata, operation: 'TRADE'}
       # memo = memo + '_TRADE'
       # from_currency -> user's mainbook to cs cashbook
       # to_currency -> cs cashbook to user's mainbook
        entries = [
            # debit user from_amount
            {
                "bookId": self.bank_account['ledger_books']['main_balance']['id'],
                "assetId": from_currency,
                "value":  str(Decimal(from_amount) * Decimal("-1"))
            },
            # credit to cs cashbook from_amount
            {
                "bookId": self.DEFAULT_CASH_BOOK_ID['id'],
                "assetId": from_currency,
                "value": str(from_amount)
            }, 
            # to_currency flow
            {
                "bookId": self.DEFAULT_CASH_BOOK_ID['id'],
                "assetId": to_currency,
                "value": str(Decimal(to_amount) * Decimal("-1"))
            },
            {
                "bookId": self.bank_account['ledger_books']['main_balance']['id'],
                "assetId": to_currency,
                "value": str(to_amount)
            }
        ]
        operation = self.ledger.post_operation(OperationType.CUSTOM, entries, memo= memo + "_TRADE", metadata={**metadata, **{"operation": "TRADE"}})
        log.info(operation)
        return operation

    def get_order_fill_status_in_bulk(self, order_id_list: list):
        log.info("---------get_Limit_order_fill_status_in_bulk----------")
        bulk_order_fill_status = []

        main_balance_order_query = []
        for order_id in order_id_list:
            main_balance_order_query.append(
                {'book_id':self.bank_account['ledger_books']['main_balance']['id'], 'metadata_filter' : {'order_id':order_id}}
            )
        main_balance_order_query_response = self.get_book_balances_bulk(main_balance_order_query)

        cashbook_balance_order_query = []
        for order_id in order_id_list:
            cashbook_balance_order_query.append(
                {'book_id': self.DEFAULT_CASH_BOOK_ID['id'], 'metadata_filter': {'order_id': order_id}}
            )
        cashbook_balance_order_query_response = self.get_book_balances_bulk(cashbook_balance_order_query)

        if not (len(cashbook_balance_order_query_response) == len(order_id_list) and len(main_balance_order_query_response) == len(order_id_list)):
            raise Exception(ErrorCodes.FETCH_ORDER_STATUS_ERROR.value)

        for counter in range(len(order_id_list)):
            order_wise_fill_status = {}
            order_wise_fill_status['fill_status'] = {}
            order_wise_fill_status['order_id'] = order_id_list[counter]
            for (key,val) in main_balance_order_query_response[counter].items():
                if Decimal(val)>0:
                    order_wise_fill_status['fill_status']['to_currency'] = key
                    order_wise_fill_status['fill_status']['to_amount'] = val
                    break
            for (key,val) in cashbook_balance_order_query_response[counter].items():
                if Decimal(val)>0:
                    order_wise_fill_status['fill_status']['from_currency'] = key
                    order_wise_fill_status['fill_status']['from_amount'] = val
                    break
            bulk_order_fill_status.append(order_wise_fill_status)

        return bulk_order_fill_status




import unittest
from accountant.client import Accountant,ErrorCodes
import random


class TestCasesOrdersAccountant(unittest.TestCase):
    def setUp(self) -> None:
        bank_account = {
            'account_id': 'test_account_id_1',
            'ledger_books': {
                'blocked_balance': {
                    'id': '2'
                },
                'main_balance': {
                    'id': '3'
                }
            }
        }
        self.accountant = Accountant(ledger_host='http://0.0.0.0:3000', bank_account=bank_account)

    def test_block_for_order(self):
        order_id = "711"
        order_type = 'limit'
        currency = 'inr'
        value = '103'
        operation = self.accountant.block_for_order(order_id, order_type, currency, value)
        self.assertIn("status", operation)
        self.assertEqual(operation['status'], 'APPLIED')

    def test_update_order_fill_status(self):
        order_id = "711"
        from_currency = 'inr'
        to_currency = 'eth'
        from_amount = '100'
        to_amount = '40'
        memo = '102'
        order_type = 'limit'
        operation = self.accountant.update_order_fill_status(order_id, order_type, from_currency, to_currency, from_amount, to_amount, memo)
        self.assertIn("status", operation)
        self.assertEqual(operation['status'], 'APPLIED')

    def test_get_order_fill_status_in_bulk(self):
        order_list = ["605"]
        fill_statuses = self.accountant.get_order_fill_status_in_bulk(order_list)
        print(fill_statuses)

    def test_unblock_for_order(self):
        order_id = "605"
        order_type = 'limit'
        operation = self.accountant.unblock_for_order(order_id, order_type)
        self.assertIn("status", operation)
        self.assertEqual(operation['status'], 'APPLIED')

    def test_flow_1(self):
        """Trying to unblock funds for an orderId for which funds were never blocked"""
        order_id = str(random.randint(1, 10000))
        order_type = 'limit'
        self.assertRaisesRegexp(Exception, ErrorCodes.INSUFFICIENT_FUNDS.value, self.accountant.unblock_for_order, order_id, order_type)

    def test_flow_2(self):
        """Trying to fill funds for an orderId for which funds were never blocked"""
        order_id = str(random.randint(1, 10000))
        from_currency = 'inr'
        to_currency = 'btc'
        from_amount = '10'
        to_amount = '1'
        memo = 'limit_101'
        order_type = 'limit'
        self.assertRaisesRegexp(Exception, ErrorCodes.INSUFFICIENT_FUNDS.value, self.accountant.update_order_fill_status,order_id,
                                order_type, from_currency, to_currency, from_amount, to_amount, memo )

    def test_flow_3(self):
        """Trying to unblock funds immediately after blocking funds"""
        order_id = str(random.randint(1, 10000))
        order_type = 'limit'
        from_currency = 'inr'
        from_amount = '100'
        self.accountant.block_for_order(order_id, order_type, from_currency, from_amount)
        operation = self.accountant.unblock_for_order(order_id, order_type)
        self.assertIn("status", operation)
        self.assertEqual(operation['status'], 'APPLIED')

    def test_flow_4(self):
        """Trying to unblock funds after blocking funds but before filling funds"""
        order_id = str(random.randint(1, 10000))
        order_type = 'limit'
        from_currency = 'inr'
        from_amount = '100'
        self.accountant.block_for_order(order_id, order_type, from_currency, from_amount)
        self.accountant.unblock_for_order(order_id, order_type)
        to_currency = 'btc'
        from_amount = '10'
        to_amount = '1'
        memo = 'limit_101'
        self.assertRaisesRegexp(Exception, ErrorCodes.INSUFFICIENT_FUNDS.value,
                                self.accountant.update_order_fill_status,
                                order_id, order_type, from_currency, to_currency, from_amount, to_amount, memo)

    def test_flow_5(self):
        """Trying to fill funds after blocking funds iteratively"""
        order_id = str(random.randint(1,10000))
        order_type = 'limit'
        from_currency = 'inr'
        from_amount = '100'
        self.accountant.block_for_order(order_id, order_type, from_currency, from_amount)
        to_currency = 'btc'
        from_amount = '10'
        to_amount = '1'
        memo = 'limit_101'
        self.accountant.update_order_fill_status(order_id, order_type, from_currency, to_currency, from_amount,
                                                 to_amount, memo)
        from_amount = '20'
        to_amount = '2'
        memo = 'limit_102'
        self.accountant.update_order_fill_status(order_id, order_type, from_currency, to_currency, from_amount,
                                                 to_amount, memo)
        from_amount = '30'
        to_amount = '3'
        memo = 'limit_102'
        operation = self.accountant.update_order_fill_status(order_id, order_type, from_currency, to_currency, from_amount,
                                                 to_amount, memo)
        self.assertIn("status", operation)
        self.assertEqual(operation['status'], 'APPLIED')

    def test_flow_6(self):
        """Trying to fill funds after blocking funds iteratively but in wrong sequence"""
        order_id = str(random.randint(1, 10000))
        order_type = 'limit'
        from_currency = 'inr'
        from_amount = '100'
        self.accountant.block_for_order(order_id, order_type, from_currency, from_amount)
        to_currency = 'btc'
        from_amount = '20'
        to_amount = '2'
        memo = 'limit_102'
        self.accountant.update_order_fill_status(order_id, order_type, from_currency, to_currency, from_amount,
                                                 to_amount, memo)
        from_amount = '10'
        to_amount = '1'
        memo = 'limit_102'
        self.assertRaisesRegexp(Exception, ErrorCodes.FILL_REQUEST_SEQUENCE_ERROR.value, self.accountant.update_order_fill_status,
                                order_id, order_type, from_currency, to_currency, from_amount, to_amount, memo)

    def test_flow_7(self):
        """Fill to_currency completely but leave some from_currency left in block book and revert it back to main book"""
        order_id = str(random.randint(1,10000))
        order_type = 'limit'
        from_currency = 'inr'
        from_amount = '100'
        self.accountant.block_for_order(order_id, order_type, from_currency, from_amount)
        to_currency = 'btc'
        from_amount = '90'
        to_amount = '10'
        memo = 'limit_102'
        self.accountant.update_order_fill_status(order_id, order_type, from_currency, to_currency, from_amount,
                                                 to_amount, memo)
        operation = self.accountant.unblock_for_order(order_id,order_type)
        self.assertIn("status", operation)
        self.assertEqual(operation['status'], 'APPLIED')

    def test_get_book(self):
        self.accountant.get_book()

    def test_trade_all_or_none(self):
        res = self.accountant.trade_all_or_none(
            from_currency="bnb",
            to_currency="inr",
            from_amount="0.02199800",
            to_amount="66.29",
            memo="060ebe62-2477-47ab-bbe0-b562e120c64d",
            metadata={}
        )
        print(res)


if __name__ == '__main__':
    unittest.main()



import time
from enum import auto
from typing import List

from jsonrpcclient.clients.http_client import HTTPClient
from jsonrpcclient.requests import Request

import logging

from jsonrpcclient.exceptions import ReceivedNon2xxResponseError

from common.cs_enum import AutoName

DEFAULT_LEDGER_BOOK_ID = "1"

log = logging.getLogger(__name__)


class OperationType(AutoName):
    TRANSFER = auto()
    CUSTOM = auto()
    VOID = auto()


class OperationStatus(AutoName):
    INIT = auto()
    PROCESSING = auto()
    APPLIED = auto()
    REJECTED = auto()


class LedgerClient:
    def __init__(self, ledger_endpoint: str, ledger_read_only_endpoint: str = None):
        self.rpc_client = HTTPClient(ledger_endpoint)
        if ledger_read_only_endpoint is not None:
            self.read_only_rpc_client = HTTPClient(ledger_read_only_endpoint)

    def _execute(self, method: str, *args, read_only_method: bool = False, retry_index=0):
        start_time = int(time.time()*1000)
        try:
            if read_only_method is True and getattr(self, "read_only_rpc_client", None) is not None:
                rpc_response = self.read_only_rpc_client.request(method, *args)
            else:
                rpc_response = self.rpc_client.request(method, *args)
        except ReceivedNon2xxResponseError as e:
            if e.code == 502 and retry_index <= 3:
                log.warning(f"Retrying 502 response", exc_info=1)
                return self._execute(method, *args, read_only_method=read_only_method, retry_index=retry_index+1)
            raise e
        end_time = int(time.time()*1000)
        log.info(f"LedgerClient: {method}: time_taken: {end_time-start_time}ms")
        return rpc_response.data.result

    def _bulk_execute(self,request_list: List, read_only_method: bool = False, retry_index=0):
        start_time = int(time.time()*1000)
        try:
            if read_only_method is True and getattr(self, "read_only_rpc_client", None) is not None:
                rpc_response = self.read_only_rpc_client.send(request_list)
            else:
                rpc_response = self.rpc_client.send(request_list)
        except ReceivedNon2xxResponseError as e:
            if e.code == 502 and retry_index <= 3:
                log.warning(f"Retrying 502 response", exc_info=1)
                return self._bulk_execute(request_list, read_only_method=read_only_method, retry_index=retry_index+1)
            raise e
        end_time = int(time.time()*1000)
        log.info(f"LedgerClient: time_taken: {end_time-start_time}ms")
        result = []
        res_id_map = {}
        for each_response in rpc_response.data:
            res_id_map[each_response.id] = each_response.result
        for each_request in request_list:
            result.append(res_id_map[each_request['id']])
        return result

    def create_book(self, name: str, min_balance: str = None, metadata: dict = None):
        book_info = dict()
        book_info["name"] = name
        book_info["metadata"] = metadata
        if min_balance:
            book_info["restrictions"] = dict()
            book_info["restrictions"]["minBalance"] = min_balance
        return self._execute("createBook", book_info)

    def get_book(self, book_id: str, timestamp: int = None):
        return self._execute("getBook", book_id, timestamp, read_only_method=True)

    def put_book(self, book_id: str, restrictions: dict):
        return self._execute("putBook", book_id, restrictions)

    def freeze_book(self, book_id: str):
        return self.put_book(book_id=book_id, restrictions={"freeze": True})

    def unfreeze_book(self, book_id: str):
        return self.put_book(book_id=book_id, restrictions={"freeze": False})

    def get_book_balances(self, book_id: str, asset_id: str = None, metadata_filter: dict = None, timestamp: int = None):
        return self._execute("getBalances", book_id, asset_id, metadata_filter, timestamp, read_only_method=True)

    def get_book_balances_bulk(self, params_list: list = []):
        request_list = []
        for params in params_list:
            if not params.get('book_id'):
                raise Exception('invalid input to get_book_balances_bulk: book_id missing')
            params['asset_id'] = params.get('asset_id',None)
            params['metadata_filter'] = params.get('metadata_filter', None)
            params['filter'] = {"toTime": params['timestamp']} if params.get('timestamp', None) else None
            request_list.append(
                Request("getBalances", params['book_id'], params['asset_id'], params['metadata_filter'], params['filter'])
            )
        return self._bulk_execute(request_list, read_only_method = True)

    def get_operation(self, operation_id: str):
        return self._execute("getOperation", operation_id, read_only_method=True)

    def get_operation_by_memo(self, memo: str):
        return self._execute("getOperationByMemo", memo, read_only_method=True)

    def get_operations(self, book_id: str, metadata_filter: dict = None):
        return self._execute("getOperations", book_id, metadata_filter, read_only_method=True)

    def post_operation(self, operation_type: OperationType, entries: List[dict], memo: str = "", metadata: dict = None):
        operation = dict()
        operation["type"] = operation_type.value
        operation["entries"] = entries
        operation["memo"] = memo
        if metadata: operation["metadata"] = metadata
        result = self._execute("postOperation", operation)
        if result["status"] == OperationStatus.REJECTED.value:
            raise Exception(result["rejectionReason"])
        return result

    def post_transfer(self, from_book_id: str, to_book_id: str, asset_id: str, value: str, memo: str = "", metadata: dict = None):
        transfer = dict()
        transfer["fromBookId"] = from_book_id
        transfer["toBookId"] = to_book_id
        transfer["assetId"] = asset_id
        transfer["value"] = value
        transfer["memo"] = memo
        if metadata: transfer["metadata"] = metadata
        result = self._execute("postTransfer", transfer)
        if result["status"] == OperationStatus.REJECTED.value:
            raise Exception(result["rejectionReason"])
        return result

    def get_diff(self, from_time: int, to_time: int):
        return self._execute("getDiff", from_time, to_time, read_only_method=True)

    def post_void(self, operation_id_list: list, memo: str, metadata: dict = None):
        void = dict()
        void["operationIds"] = operation_id_list
        void["memo"] = memo
        void["metadata"] = metadata or {}
        result = self._execute("postVoid", void)
        if result["status"] == OperationStatus.REJECTED.value:
            raise Exception(result["rejectionReason"])
        return result
